#!/usr/bin/perl
#  application: SQL QuickRegister Password Management System
#       script: genlib.pl
#      Version: March/2002 1.0
#       author: David Jourard
#               Bytes Interactive
#               http://www.bytesinteractive.com
#               cgi@bytesinteractive.com
# requirements: perl 5 Unix Version and mySQL
#
#
#  description: general functions
# ***************************************************************************
# SQL QuickRegister
# COPYRIGHT NOTICE
# Copyright 2002 David Jourard  All Rights Reserved.
#
# This program cannot be used till one reads and agrees the terms of the license
# defined in the file license.txt which comes with the software.
#*****************************************************************************
package genlib;

use strict;
use CGI::Carp;
use vars qw($gvar $ovar $cvar %config $debug %QEVar @referers %c  @othertemplates);

use Time::Local;
use Digest::MD5;

# Uncomment (take away the #) if using for sending e-mail
#use Net::SMTP;


$debug  = $main::debug;

if($debug) { print "Accessing genlib.pl ...\n"; }

##############################################################################
# Global variables need to set

# for CheckUrl - what domains allowed to use the script
@referers = ('yourdomain.com', 'your.ip.add.ress');

# for ParseforTag - defines the XML prefix
$gvar->{'XMLPrefix'} = 'QR';

# To make case insensitive the XML tags need to define how the variable is
# defined in the script wrt case - eg. in the script its $XmlName on the LHS
# The index is in upper case.
#$QEVar{'XMLNAME'} = 'XmlName';

##############################################################################

sub SetStyleLine {
   my($file) = $_[0];
   my($StyleLine)='';
   # Style 2 - custom
   open(HITM, $file) || &SysError("SetStyleLine", "Cannot open $file");
   while(<HITM>) {
      # Concatenate lines into one variable
      $StyleLine .= $_;
   }
   close(HITM);
   return ($StyleLine);
}

sub FrmtHtmlList {
   my($r_html) = $_[0];
   my($CopyStyleLine) = $_[1];

   $CopyStyleLine = &ParseForVars($CopyStyleLine);
   $CopyStyleLine = &ParseForTag($CopyStyleLine);
   $$r_html .= $CopyStyleLine;

}

sub BldHref {
   my($text) = $_[0];
   my($href) = $_[1];
   my($rtrn);
   $rtrn = qq(<a href="$href"> $text </a>);
   return( $rtrn );
}

sub ParseTxt {

   my($varlen, $ni, $offset, $info, $charcut, $LINELIMIT, $ln);
   my(@lines);

   $info=$_[0];

   $LINELIMIT=60;
   #Edit comment lines to 50 characters
   $offset=$LINELIMIT;
   $ni=1;
   $varlen=length($info);

   $info =~ s/\n//g;
   $info =~ s/\x0D//g;

   if ($varlen > $LINELIMIT) {
      # Take out all the linefeeds

      #Parse if one linefeed or less
      while($varlen>$LINELIMIT) {
           #Parse

           $charcut=substr($info, $offset, 1);
           if ($charcut eq " ") {
              #Not Middle of word break up
              $lines[$ni]=substr($info,0,$offset+1);
              $lines[$ni] = $lines[$ni]."\n";
              $ni++;
              $varlen=$varlen-$offset+1;
              $info=substr($info,$offset+1, $varlen);

              #Reset offset
              $offset=$LINELIMIT;
           }
           elsif ($offset == 0) {
              #Entire line is made up of characters
              $lines[$ni]=substr($info,0,$LINELIMIT+1);
              $lines[$ni] = $lines[$ni]."\n";
              $ni++;
              $varlen=$varlen-$LINELIMIT+1;
              $info=substr($info,$LINELIMIT+1, $varlen);

              #Reset offset
              $offset=$LINELIMIT;
           }
           else {
             # Decrease Offset
             $offset=$offset-1;
           }
      }
      if (length($info)>0) {
         # Don't forget last line
         $lines[$ni]=$info;
         $lines[$ni] = $lines[$ni]."\n";
      }
   }
   else {
      # One line less than linelimit
      $lines[0] = $info;
      $lines[0] = $lines[0]."\n";
   }

   $ni=0;
   foreach $ln (@lines) {
      if ($ln =~ /\./) {
         $ln =~ s/\./\.  /g;
         $lines[$ni] = $ln;
      }
      $ni++;
   }

   return @lines;
}

sub ParseData {
   my($name, $value, $pair, $dummy, $buffer, $meth, $qstr, $type, $clen);
   my(@pairs, %VALUES, $nobytesread);

   # Security Hole 1 - prevent script access from outside this server
   # Check Referring URL
   &check_url;

   $clen = $ENV{'CONTENT_LENGTH'};
   $qstr = $ENV{'QUERY_STRING'};
   $type = $ENV{'CONTENT_TYPE'};
   $meth = $ENV{'REQUEST_METHOD'};

   if ($meth eq 'POST') {
      $nobytesread = read(STDIN, $buffer, $clen, 0);
      if ($nobytesread != $clen) {
         &InitializeOut();
         print qq(<html>\n <head>\n);
         print qq(</head>\n <body>);
         print qq(<br><br><br><center><h2> System Error - Read $nobytesread bu
needed $clen </h2>);
         print qq(</body>\n </html>);
         exit;
      }
      if ($qstr) {
         # Append GET
         $buffer = join('', $buffer, "\&", $qstr);
      }
   }
   elsif ($meth eq 'GET') {
      # Read in GET
      $buffer = $ENV{'QUERY_STRING'};
   }
   elsif ($qstr) {
      $buffer = $ENV{'QUERY_STRING'};
   }

   # Security Hole 2 - remove relative path access
   $buffer =~ s/(\/\.\.\/)*//g;
   $buffer =~ s/(\.\.\/)*//g;
   $buffer =~ s/(\.\/)*//g;
   $buffer =~ s/(\.\.)*//g;

   # Security Hole 3 - Escape insecure Unix characters
   $buffer =~ s/([;<>\*\|`\$!#\(\)\[\]\{\}:'"]@)/\\$1/g;

   # Security Hole 4 - Remove server side includes
   $buffer =~ s/<!--(.|\n)*-->//g;

   if ($debug) { print "ParseDatabuffer=$buffer\n"; }

   # Slurp & Parse form data
   @pairs = split (/&/, $buffer);
   foreach $pair (@pairs)
   {
      ($name, $value) = split(/=/, $pair);
      $name =~ tr/+/ /;
      $value =~ tr/+/ /;
      $name =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;
      $value =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;


      #Trap for image variables
      if ($name =~ /img_/) {
         ($name, $dummy) = split(/\./, $name);
         ($dummy, $name, $value) = split(/_/, $name);
      }
      if ($debug) { print "name=$name\n"; }
      if ($debug) { print "value=$value\n"; }

      if ($name eq 'states' || $name eq 'counties') {
         # Track multiple selects by using : as a delimiter
         if ($VALUES{$name}) {
            $VALUES{$name} = join(':', $VALUES{$name}, $value);
         }
         else {
            $VALUES{$name} = $value;
         }
      }
      else {
         $VALUES{$name} = $value;
      }
   }
   return (%VALUES);
}

sub RtrnErrorMsg {
   my($RtrnFile) = $_[0];

   #Build error message with HTML
   &InitializeOut();
   &RdOutFile($RtrnFile);
}

sub RtrnPage {
   my($RtrnFile) = $_[0];
   &InitializeOut();
   &RdOutFile($RtrnFile);
}


sub ChkEmptyFields {

   my($rh_in1) = $_[0];
   my($Fields) = $_[1];
   my($strng) = "";
   my(@empty_fields) = ();

   if ($debug) { print STDOUT "validating";}
   @empty_fields = &ValStatic($rh_in1, $Fields);
   if ($debug) { print STDOUT "no empty field=$#empty_fields\n";}
   if ($#empty_fields >= 0) {
      $strng = &BldAndList(\@empty_fields, 0);
   }
   return($strng);

}

sub ValStatic {
   my($rh_in) = $_[0];
   my($Fields) = $_[1];
   my($ni) = 0; my($fieldnm)=''; my($fieldesc)='';
   my($fieldtp)=''; my($fieldnodef)=''; my($field)='';;
   my(@fieldlist)=();    my(@vldfldnm)=();  my(%vldflddesc)=();
   my(%vldfldtype)=();   my(@empty_fields) = ();

   if ($debug) { print STDOUT "checking fields\n";}
   (@fieldlist) = split(/\^/, $Fields);
   $ni=0;
   foreach $field (@fieldlist) {
      ($fieldnm, $fieldesc, $fieldtp) = split(/\?/,$field);
      $vldfldnm[$ni++]=$fieldnm;
      $vldflddesc{$fieldnm}=$fieldesc;
      $vldfldtype{$fieldnm}=$fieldtp;
      if ($debug) { print STDOUT "field=$field\n";}
      if ($debug) { print STDOUT "fieldnm=$fieldnm\n";}
      if ($debug) { print STDOUT "fielddes=$fieldesc\n";}
      if ($debug) { print STDOUT "fieldtp=$fieldtp\n";}
   }

   $ni=0;
   foreach $field (@vldfldnm) {
      if ($debug) { print STDOUT "field=$field\n";}
      if ($vldfldtype{$field}) {
         # Select fields to be tested
         if (${$vldfldtype{$field}} ) {
            if ( !$$rh_in{$field} ) {
              if ($debug) { print STDOUT "empty fieldesc=$vldflddesc{$field}\n";}
              $empty_fields[$ni++]=$vldflddesc{$field};
            }
            ${$vldfldtype{$field}} = 0;
         }
      }
      else {
         if ($debug) { print STDOUT "fld value=$$rh_in{$field}\n";}
         if ( !$$rh_in{$field} ) {
           if ($debug) { print STDOUT "empty fieldesc=$vldflddesc{$field}\n";}
           $empty_fields[$ni++]=$vldflddesc{$field};
         }
      }
   }
   return @empty_fields;
}

sub BldAndList {

   my($ra_ary) = $_[0];
   my($lLineBreak) = $_[1];
   my($strng) = ''; my($fldmsg)=''; my($ji)=0;

   $ji = 0;
   foreach $fldmsg (@$ra_ary) {
      if ($ji == 0) {
         $strng = $fldmsg;
      }
      elsif ($ji == ($#$ra_ary)) {
         if ($ji == 1) {
            #just two
            if ($lLineBreak) { $strng = join(' and<br> ',$strng,$fldmsg);}
            else { $strng = join(' and ',$strng,$fldmsg); }
         }
         else {
            #more than two
            if ($lLineBreak) { $strng = join(', and<br> ',$strng,$fldmsg); }
            else { $strng = join(', and ',$strng,$fldmsg); }
         }
      }
      else {
         if ($lLineBreak) { $strng = join(',<br> ',$strng,$fldmsg); }
         else { $strng = join(', ',$strng,$fldmsg); }
      }
      $ji++;
   }

   return $strng;
}


sub ValEmail {
   my($email) = $_[0];
   my($lRtVl) = 0;
   my($fst, $snd, $er, $er2);


   #######################################
   #      Validate email address
   #######################################
   # 1.
   $er2 = ( $email =~ /(\[|;|>|<|&|\*|`|\\|\||])/ );
   if ( $er2 ) {
      return( "The E-mail address $email contains invalid characters");
   }


   # 2.
   $er = ($email =~ /\@/);
   if (!$er) {
     # At least one name entry must be filled
     return( "Your e-mail address is missing the \@ character.");
   }


   # 3.
   ($fst,$snd) = split(/\@/, $email );
   if ($email) {
      if (length($snd) == 0 || length($fst) == 0)  {
       # At least one name entry must be filled
        return( "Your e-mail address is incomplete. It requires information on both sides of the \@ character.");
      }
   }

   return($lRtVl);
}

sub InitializeOut {
   print STDOUT "Content-type: text/html\n\n";
}

sub RdOutFile {
   my($file) = $_[0];
   my($outhandle) = $_[1];
   my($r);

   if ($debug) { print "RdOutFile ...\n"; }
   #if ($debug) { print "file1=$file\n"; }

   &ChkFile(\$file);
   #if ($debug) { print "file2=$file\n"; }

   #if ($debug) {
   #   my($key) = '';
   #   foreach $key (keys %{$main::ovar}) {
   #      print " ovar in rdo $key = $main::ovar->{$key}\n";
   #   }
   #}
   open(DXYZ, "$file") || &SysError("RdOutFile", "Cannot open $file");
   #if ($debug) { print "r=$r\n"; }
   while ( <DXYZ> ) {
      $_ = &ParseForVars($_);
      $_ = &ParseForTag($_);
      if (!$outhandle) {
         print STDOUT $_;
      }
      else {
         print $outhandle $_;
      }
   }
   close(DXYZ);
}

sub ParseForTag {
  my($fileline) = $_[0];
  my($ln, $line, $nk, $ni, $word, $up, $ch);
  my(@vars);
  my($var, $lnID, $LeadChrs, $ptr, $idchrs, $ncnt);

  #if ($debug) { print "ParseForTag .....\n"; }

  $lnID = 2;
  #$LeadChrs = "QE";

  $LeadChrs = $gvar->{'XMLPrefix'};

  $ln = length($fileline);
  $ptr = 0;
  $word = "";
  $nk=0;
  #$SafetyValve=1000;

  #if ($main::debug) { print "LeadChrs=$LeadChrs\n"; }
  #if ($main::debug) { print "line=$line\n"; }


  while ( $ptr < $ln+1 ) {
     $ch = substr($fileline, $ptr++, 1);

     if ($ch eq '<') {
        # Inside Tag - find ours
        while (1) {
           $idchrs = substr($fileline, $ptr, $lnID);
           $idchrs = "\U$idchrs";
           if ($idchrs eq $LeadChrs) {
              #if ($main::debug) { print "line2=$line\n"; }
              # Found it extract
              $ptr+=3;
              $ch="";
              $ncnt=0;
              while( $ch ne ">" ) {
                 $ch = substr($fileline, $ptr++, 1);
                 if ($ch eq '/') {
                    $ch = '\/';
                 }
                 $word = join("", $word, $ch);

                 #if ( $ncnt > ($ln+$SafetyValve) ) { return (0); }
                 $ncnt++;
              }
              $word =~ s/\s*//g;
              chop($word); # take off the >
              $vars[$nk++] = $word;
              $word='';
           }
           last; # next set of characters
        }
     }
  }

  # Remove case sensitivity by making all upper
  foreach $var (@vars) {
     $var =~ s/\s*\$//;
     # Remove case sensitivity by making all upper
     #if ($main::debug) { print "var=$var\n"; }
     $var = "\U$var";
     #if ($main::debug) { print "qevar=$QEVar{$var}\n"; }
     $var = $main::QEVar{$var};
     if ($main::ovar->{$var} || $main::ovar->{$var} eq '0') {
        $fileline =~ s/\<$LeadChrs\s*\$$var>/$main::ovar->{$var}/i;
     }
     else {
       $fileline =~ s/\<$LeadChrs\s*\$${var}>//i;
     }

  }
  return($fileline);

}

sub ParseForVars {
  my($fileline) = $_[0];
  my($ln, $line, $nk, $ni, $word, $up, $ch);
  my(@vars);
  my($var, $lnID, $LeadChrs, $ptr, $idchrs, $ncnt);

  #if ($debug) { print "ParseForVars .....\n"; }

  $lnID = 1;
  $LeadChrs = "\$";
  $ln = length($fileline);
  $ptr = 0;
  $word = "";
  $nk=0;
  #$SafetyValve=1000;
  my $delim1 = "\%";
  my $delim2 = "\%";
  my $StartCnt = 1;

  #if ($debug) { print "fileline=$fileline\n"; }

  while ( $ptr < $ln+1 ) {
     $ch = substr($fileline, $ptr++, 1);

     if ($ch eq $delim1) {
        #if ($debug) { print "fileline=$fileline\n"; }
        # Inside Tag - find ours
        while (1) {
           $idchrs = substr($fileline, $ptr, $lnID);
           $idchrs = "\U$idchrs";
           if ($idchrs eq $LeadChrs) {
              #if ($debug) { print "found one ...\n"; }
              # Found it extract
              $ptr+=$StartCnt;
              $ch="";
              $ncnt=0;
              #while( $ch ne ">" ) {
              while( $ch ne $delim2) {
                 $ch = substr($fileline, $ptr++, 1);
                 if ($ch eq '/') {
                    $ch = '\/';
                 }
                 $word = join("", $word, $ch);

                 #if ($debug) { print "word=$word...\n"; }
                 #if ( $ncnt > ($ln+$SafetyValve) ) { return (0); }
                 $ncnt++;
              }
              $word =~ s/\s*//g;
              chop($word); # take off the >
              $vars[$nk++] = $word;
              $word='';
           }
           last; # next set of characters
        }
     }
  }

  foreach $var (@vars) {
     $var =~ s/\s*\$//;
     #if ($debug) { print "var=$var...\n"; }
     #if ($debug) { print "var=${var}...\n"; }
     #if ($debug) { print "var=$${var}...\n"; }

     #if ($debug) { print "ovar=$main::ovar->{$var}...\n"; }

     if ($main::ovar->{$var} || $main::ovar->{$var} eq '0') {
        $fileline =~ s/\%\$${var}\%/$main::ovar->{$var}/;
     }
     else {
        $fileline =~ s/\%\$${var}\%//;
     }
  }

  return($fileline);

}


sub FixTime {
   my($h) = $_[0];
   my($m) = $_[1];
   my($s) = $_[2];

   if (length($h) == 1) {
      $h = '0'.$h;
   }
   if (length($m) == 1) {
      $m = '0'.$m;
   }
   if (length($s) == 1) {
      $s = '0'.$s;
   }

   return(join(".",$h,$m,$s));
}
sub FixDate {
   my($dt);
   my($a)=$_[0];
   my($b)=$_[1];
   my($c)=$_[2];
   my($sep)=$_[3];

   if (length($a) == 1) {
      $a = '0'.$a;
   }
   if (length($b) == 1) {
      $b = '0'.$b;
   }
   if (length($c) == 1) {
      $c = '0'.$c;
   }
   $dt = join($sep,$a,$b,$c);

   return($dt);
}

sub FixSqlDate {
   my($dt);
   my($y)=$_[0];
   my($m)=$_[1];
   my($d)=$_[2];
   my($sep)=$_[3];

   if (length($d) == 1) {
      $d = '0'.$d;
   }
   if (length($m) == 1) {
      $m = '0'.$m;
   }
   $dt = join($sep,$y,$m,$d);

   return($dt);
}

sub CreateDate {
   my($r_time) = $_[0];
   my($r_longdate) = $_[1];
   my($r_shortdate) = $_[2];
   my($r_shortdate2) = $_[3];
   my($r_sqldatefrmt) = $_[4];
   my($r_longtime) = $_[5];

   my($year)=''; my(@timedata) = (); my($mon)=''; my($min)='';
   my($decade)='20';

   my(%weekday, %month);

   my $thetime = time;

   my $ch_tm = 0;

   # Move time back or forward in seconds
   if ($main::gvar->{'AdjustTimeInHours'}) {
      $ch_tm = 3600 * $main::gvar->{'AdjustTimeInHours'};
   }
   $thetime = $thetime + $ch_tm;

   $ch_tm = 0;
   if ($main::gvar->{'AdjustTimeInMinutes'}) {
      $ch_tm = 60 * $main::gvar->{'AdjustTimeInMinutes'};
   }
   $thetime = $thetime + $ch_tm;

   (@timedata) = localtime($thetime);
#$sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = localtime(time);

   $year = substr($timedata[5], 1, 2);

   %weekday = ("0", "Sunday", "1", "Monday", "2", "Tuesday", "3", "Wednesday","4", "Thursday", "5", "Friday","6", "Saturday",);
   %month = ("0", "January", "1", "February", "2", "March", "3", "April","4", "May", "5", "June","6", "July", "7", "August", "8", "September","9", "October", "10", "November", "11", "December",);

   my $fullyear = join('', $decade, $year);

   $$r_longdate = join("", $weekday{$timedata[6]}, , " ", $month{$timedata[4]}, " ", $timedata[3], ", $fullyear");

   $mon= $timedata[4]+1;
   $$r_shortdate = &FixDate($timedata[3], $mon, $year, '.');

   $$r_shortdate2 = join(".", $timedata[3], $mon, $year);

   $$r_sqldatefrmt = &FixSqlDate($fullyear, $mon, $timedata[3],"-");


   my $hour = $timedata[2];
   my $minute = $timedata[1];

   $$r_time = &FixTime($hour, $minute, $timedata[0]);

   $min = $minute;
   if (length($min) == 1) {
      $min = '0'.$min;
   }


   if ($hour <= 11) {
      $$r_longtime = join("", $hour, ":", $min, " AM");
   }
   elsif ($hour == 12) {
      $$r_longtime = join("", $hour, ":", $min, " PM");
   }
   else {
      $hour = $hour - 12;
      $$r_longtime = join("", $hour, ":", $min, " PM");
   }

}

sub SendUnixMail {
   my($EmailTo) = $_[0];
   my($EmailFrom) = $_[1];
   my($EmailFromAdd) = $_[2];
   my($EmailSubject) = $_[3];
   my($EmailCC) = $_[4];
   my($EmailTxt) = $_[5];

   my($FromAdd, $r);
   if ($debug) { print "SendUnixMail ...\n"; }
   if ($debug) { print "   EmailTo=$EmailTo ..\n"; }
   if ($debug) { print "   EmailFrom=$EmailFrom ..\n"; }
   if ($debug) { print "   EmailFromAdd=$EmailFromAdd ..\n"; }
   if ($debug) { print "   EmailSubject=$EmailSubject ..\n"; }


   $FromAdd = join("", "\"", $EmailFrom, "\" ", "<", $EmailFromAdd, ">");

   if ($debug) { print "  FromAdd = $FromAdd ...\n"; }
   $r = open(MAIL, "| $main::gvar->{'MailPath'} -t") || &SysError("SendUnixMail", "Cannot open sendmail. Check the parameter MailPath under Edit Configuration Parameters", 1);
   if ($debug) { print "sendmail_return_vl = $r\n"; }
   print MAIL "To: $EmailTo\n";
   print MAIL "From: $FromAdd\n";
   if ($EmailCC) {
      print MAIL "Cc: $EmailCC\n";
   }
   print MAIL "Subject: $EmailSubject\n\n";
   &RdOutFile($EmailTxt, \*MAIL);
   close(MAIL);
   if ($debug) { print "end_of_SendUnixMail ...\n"; }
}


sub SendSmtpMail {
   my($EmailTo) = $_[0];
   my($EmailFrom) = $_[1];
   my($EmailFromAdd) = $_[2];
   my($EmailSubject) = $_[3];
   my($EmailCC) = $_[4];
   my($EmailTxt) = $_[5];

   my($FromAdd) = '';
   if ($debug) { print "SendSmtpMail ...\n"; }
   if ($debug) { print "   EmailTo=$EmailTo ..\n"; }
   if ($debug) { print "   EmailFrom=$EmailFrom ..\n"; }
   if ($debug) { print "   EmailFromAdd=$EmailFromAdd ..\n"; }
   if ($debug) { print "   EmailSubject=$EmailSubject ..\n"; }


   $FromAdd = join("", "\"", $EmailFrom, "\" ", "<", $EmailFromAdd, ">");

   # Email file check if exists
   &ChkFile(\$EmailTxt);

   my $MailHost = $main::gvar->{'MailHost'};

   #my $smtp = Net::SMTP->new($MailHost,  Hello => $ENV{'SERVER_NAME'});
   my $smtp = Net::SMTP->new($MailHost);
   my $is_sent = $smtp->mail($EmailFromAdd);
   if (!$is_sent) {
      &SysError("SendSmtpMail", "SMTP E-mail Failed.  Try using SMTP Module", 1);
      exit;
   }
   $smtp->to($EmailTo);
   $smtp->data;
   $smtp->datasend("From: $FromAdd\n");
   $smtp->datasend("Subject: $EmailSubject\n");
   $smtp->datasend("\n");

   open(DXYZ, "$EmailTxt") || &SysError("SendSmtpMail","Cannot open $EmailTxt");
   #if ($debug) { print "r=$r\n"; }
   while ( <DXYZ> ) {
      $_ = &ParseForVars($_);
      $_ = &ParseForTag($_);
      $smtp->datasend($_);
   }
   close(DXYZ);

   $is_sent=$smtp->dataend;
   if (!$is_sent) {
      &SysError("SendSmtpMail", "SMTP E-mail Failed.  Try using sendmail Module");
      exit;
   }

   $smtp->quit;

   if ($debug) { print "end_of_SendSmtpMail ...\n"; }
}



sub PrepDataStuff {
   my($ra_data_var) = $_[0];
   my($rh_VALUES) = $_[1];
   my($step, $fld);
   my($DataStuff);

   $DataStuff='';
   foreach $fld (@$ra_data_var) {
      $step = join("=", $fld, $$rh_VALUES{$fld});
      $DataStuff = join("&", $DataStuff, $step);

   }
   $DataStuff =~ s/&//;
   $DataStuff = qq(<input type="hidden" name="DataStuff" value="$DataStuff">);

   return($DataStuff);
}

sub CalcDate {
      my($today) = $_[0];
      my($add_days) = $_[1];
      my($datetype) = $_[2];
      my($diff, $rtrn_date);

      my($dy, $mth, $year) = split(/\./, $today);

      my(%mths) = (4=>30, 6=>30, 9=>30, 11=>30, 1=>31, 3=>31, 5=>31, 7=>31,8=>31, 10=>31, 12=>31, 2=>28);
      my(%month) = ("1", "January", "2", "February", "3", "March", "4", "April","5", "May", "6", "June","7", "July", "8", "August", "9", "September","10", "October","11", "November", "12", "December",);

      my($nxtdy, $nxtmth);

      $nxtmth = $month{$mth};
      $nxtdy = $dy+$add_days;

      # 25,32,1  26,33,2
      $diff = $nxtdy - $mths{$mth};
      #  25 ---> 1, 26 ----> 2, 27,3 28,4 29,5 30,6 31,7 1,8

      if ($diff >= 1) {
         $nxtdy=$diff;
         if ($mth < 12) {
            $mth++;
            $nxtmth = $month{$mth};
         }
         else {
            $mth = 1;
            $nxtmth = $month{'1'};
            $year = $year + 1;
            if ($year <=9 ) { $year = join("", '0', $year); }
         }
      }


      if ($datetype == 1){

         $rtrn_date = join("", $nxtmth, " $nxtdy,", " 20", $year);
      }
      else {
         if ($mth <=9 ) { $mth = join("", '0', $mth); }
         if ($nxtdy <=9 ) { $nxtdy = join("", '0', $nxtdy); }
         $rtrn_date = join("/", $mth, $nxtdy, $year);
      }

      return $rtrn_date;
}


sub send_mail
{
   my($fromsender,$recipients,$subject,$message_file,$cc)=@_;
   my($line, $rtrn);
   # Send mail using blat and the system command ############

   $line = join(" ", $gvar->{'BlatPath'}, $message_file, "-t $recipients","-f $fromsender", "-s \"$subject\"");


   if ($debug) { print "line=$line\n"; }

   $rtrn=system($line);

   if ($debug) { print "rtrn=$rtrn\n"; }
}

sub RoundOff {
  return( sprintf "%.2f", $_[0] );
}

sub PadL {

   my($pdvl) = $_[0];
   my($pdln) = $_[1];
   my($str) = $_[2];
   my($lnfld, $lnstr, $totdiff);

   $lnstr = length($str);
   if ($pdln > $lnstr) {
      $totdiff=$pdln-$lnstr;
      while ($totdiff != 0) {
         #$str .= $pdvl;
         $str = join("", $pdvl, $str);
         $totdiff--;
      }
   }
   return $str;

}

sub WrError {
   my($function) = $_[0];
   my($message) = $_[1];

   if ($debug) { print "WrError ....\n"; }

   my($gvar2) = {};
   &CreateDate(\$gvar2->{'time'}, \$gvar2->{'longdate'},\$gvar2->{'date'},\$gvar2->{'shortdate2'}, \$gvar2->{'longtime'});
   open(ER, ">>$main::config{'DataPath'}/error/$gvar->{'ErrorLog'}") || &died(0, $gvar->{'ErrorLog'});
   print ER join(':',$function, $gvar2->{'longdate'},$gvar2->{'time'}, $message );
   print ER "\n";
   close(ER);
}

sub SysError {

    my($function) = $_[0];
    my($errormsg) = $_[1];
    my($donotuseemail) = $_[2];


    my($SystemError) = '';  my($mailpth) = '';
    if ($debug) { print "SysError ....\n"; }

    my($gvar2) = ();

    &CreateDate(\$gvar2->{'time'}, \$gvar2->{'longdate'},\$gvar2->{'date'},\$gvar2->{'shortdate2'}, \$gvar2->{'longtime'});

    # E-mail the admin a system error
    if (!$main::gvar->{'AdminEmail'}) {
       $errormsg .= qq(<br> Also AdminEmail needs to be set under Edit Configuration Parameters);
       $donotuseemail = 1;
    }


    my $rtrn = 0;
    if (!$donotuseemail) {
    if ($main::gvar->{'MailType'} == 0) {

       if ($debug) { print "Sendmail e-mail ...\n"; }

       $mailpth = join("", "\|", $main::gvar->{'MailPath'}," -t");
       $rtrn = open(MAIL, $mailpth);
       if ($rtrn) {
          print MAIL "To: $main::gvar->{'AdminEmail'}\n";
          print MAIL "From: $main::gvar->{'SystemName'}\n";
          print MAIL "Subject: $main::gvar->{'ErrorSubject'}\n\n";

          print MAIL "        Date: $gvar2->{'longdate'}\n";
          print MAIL "        Time: $gvar2->{'longtime'}\n";
          print MAIL "    Function: $function\n";
          print MAIL "System Error: $errormsg\n";
       }
       else {
         $errormsg .= qq(<br> Also MailPath may not be set correctly under Edit Configuration Parameters rtrn=$rtrn ml=$mailpth);
       }
       close(MAIL);
    }
    else {

       if ($debug) {
          print "SMTP e-mail ...\n";
          open(STDERR, ">&STDOUT");
       }

       my $EmailTo = $main::gvar->{'AdminEmail'};
       my $EmailFrom = $main::gvar->{'AdminEmail'};
       my $FromAdd = qq("$main::gvar->{'SystemName'}" <$EmailTo>);
       my $EmailSubject = $main::gvar->{'ErrorSubject'};
       my $MailHost = $main::gvar->{'MailHost'};
       my $longdate = $gvar2->{'longdate'};
       my $longtime = $gvar2->{'longtime'};

       my $smtp = '';
       if ($debug) {
          $smtp = Net::SMTP->new($MailHost, Debug=>1, Hello => $ENV{'SERVER_NAME'});
       }
       else {
          #$smtp = Net::SMTP->new($MailHost, Hello => $ENV{'SERVER_NAME'});
          $smtp = Net::SMTP->new($MailHost);
       }
       $smtp->mail($EmailFrom);
       $smtp->to($EmailTo);
       $smtp->data;
       $smtp->datasend("From: $FromAdd\n");
       $smtp->datasend("Subject: $EmailSubject\n");
       $smtp->datasend("\n");
       $smtp->datasend("Date: $longdate\n");
       $smtp->datasend("Time: $longtime\n");
       $smtp->datasend("Function: $function\n");
       $smtp->datasend("System Error: $errormsg\n");
       $smtp->dataend;
       $smtp->quit;

    }
    }

    #Build error message with HTML
    $QEVar{'SYSTEMERROR'} = 'SystemError';
    $SystemError .= "Date: $gvar2->{'longdate'}<br>";
    $SystemError .= "Time: $gvar2->{'longtime'}<br>";
    $SystemError .= "System Error:  $errormsg<br>";

    &InitializeOut();
    print qq(<html>\n <head>\n);
    print qq(</head>\n <body>);
    print qq(<br><br><br>$SystemError);
    print qq(</body>\n </html>);

    # Write out the system error to log file
    open(ER, ">>$main::config{'DataPath'}/error/$main::gvar->{'ErrorLog'}") || &died(0, $main::gvar->{'ErrorLog'});
    print ER join(':',$function, $gvar2->{'longdate'},
$gvar2->{'longtime'}, $errormsg );
    print ER "\n";
    close(ER);
    if ($debug) { print "System Error: $errormsg ...\n"; }
    exit;
}


sub SysSqlError {

    my($function) = $_[0];
    my($errormsg) = $_[1];
    my($sqlcmd) = $_[2];
    my($dbierr) = $_[3];
    my($dbierrstr) = $_[4];

    my($SystemError) = '';  my($mailpth) = '';    my($gvar2) = ();
    if ($debug) { print "SysSqlError ....\n"; }


    &CreateDate(\$gvar2->{'time'}, \$gvar2->{'longdate'},\$gvar2->{'date'},\$gvar2->{'shortdate2'}, \$gvar2->{'longtime'});

    # E-mail the admin a system error
    if ($main::gvar->{'MailType'} == 0) {
       $mailpth = join("", "\|", $main::gvar->{'MailPath'}," -t");
       open(MAIL, $mailpth) || &SysError("SysError","$errormsg and Cannot open sendmail to send error: Configure MailPath in config.txt");
       print MAIL "To: $main::gvar->{'AdminEmail'}\n";
       print MAIL "From: $main::gvar->{'SystemName'}\n";
       print MAIL "Subject: $main::gvar->{'ErrorSubject'}\n\n";

       print MAIL "         Date: $gvar2->{'longdate'}\n";
       print MAIL "         Time: $gvar2->{'longtime'}\n";
       print MAIL "     Function: $function\n";
       print MAIL " System Error: $errormsg\n";
       print MAIL "   DBI::error: $dbierr\n";
       print MAIL "DBI::errorstr: $dbierrstr\n";
       if ($sqlcmd) {
          print MAIL " SQL Command: $sqlcmd\n";
       }
       close(MAIL);
    }
    else {
       if ($debug) {
          print "SMTP e-mail ...\n";
          open(STDERR, ">&STDOUT");
       }

       my $EmailTo = $main::gvar->{'AdminEmail'};
       my $EmailFrom = $main::gvar->{'AdminEmail'};
       my $FromAdd = qq("$main::gvar->{'SystemName'}" <$EmailTo>);
       my $EmailSubject = $main::gvar->{'ErrorSubject'};
       my $MailHost = $main::gvar->{'MailHost'};
       my $longdate = $gvar2->{'longdate'};
       my $longtime = $gvar2->{'longtime'};

       my $smtp = '';
       if ($debug) {
          $smtp = Net::SMTP->new($MailHost, Debug=>1, Hello => $ENV{'SERVER_NAME'});
       }
       else {
          #$smtp = Net::SMTP->new($MailHost, Hello => $ENV{'SERVER_NAME'});
          $smtp = Net::SMTP->new($MailHost);
       }
       $smtp->mail($EmailFrom);
       $smtp->to($EmailTo);
       $smtp->data;
       $smtp->datasend("From: $FromAdd\n");
       $smtp->datasend("Subject: $EmailSubject\n");
       $smtp->datasend("\n");
       $smtp->datasend("Date: $longdate\n");
       $smtp->datasend("Time: $longtime\n");
       $smtp->datasend("Function: $function\n");
       $smtp->datasend("System Error: $errormsg\n");
       $smtp->datasend("DBI::error: $dbierr\n");
       $smtp->datasend("DBI::errorstr: $dbierrstr\n");
       if ($sqlcmd) {
          $smtp->datasend("SQL Command: $sqlcmd\n");
       }
       $smtp->dataend;
       $smtp->quit;
    }


    #Build error message with HTML
    $QEVar{'SYSTEMERROR'} = 'SystemError';
    $SystemError .= "Date: $gvar2->{'longdate'}<br>";
    $SystemError .= "Time: $gvar2->{'longtime'}<br>";
    $SystemError .= "System Error:  $errormsg<br>";
    if ($sqlcmd) {
       $SystemError .= "SQL statement:  $sqlcmd<br>";
    }
    $SystemError .= "DBI::err:  $dbierr<br>";
    $SystemError .= "DBI::errstr:  $dbierrstr<br>";

    &InitializeOut();
    print qq(<html>\n <head>\n);
    print qq(</head>\n <body>);
    print qq(<br><br><br>$SystemError);
    print qq(</body>\n </html>);


    if ($sqlcmd) {
       $errormsg .= join(' ', $errormsg, $sqlcmd, $dbierr, $dbierrstr);
    }
    else {
       $errormsg .= join(' ', $errormsg, $dbierr, $dbierrstr);
    }
    # Write out the system error to log file


    open(ER, ">>$main::config{'DataPath'}/error/$main::gvar->{'SqlErrorLog'}") || &died(0, $main::gvar->{'SqlErrorLog'});
    print ER join(':', $function, $gvar2->{'longdate'},
$gvar2->{'longtime'}, $errormsg );
    print ER "\n";
    close(ER);

    if ($debug) { print "System Error: $errormsg ...\n"; }
    exit;
}

sub OpenParam {
   my($file) = $_[0];
   my($rh_var) = $_[1];

   my($paramname, $paramval);
   if ($debug) { print "OpenParam ....\n"; }

   if ($debug) { print "   file=$file\n"; }

   open(SYS, "$file") || &SysError("OpenParam", "Cannot open $file");
   while (<SYS>) {
      chop($_);
      $_ || next;
      unless ($_ =~ /#/) {
         ($paramname, $paramval) = split("=", $_, 2);
         $rh_var->{$paramname} = $paramval;
         if ($debug) { print " $paramname = $rh_var->{$paramname}\n"; }
      }
   }
}

sub TakeOutSpCh {
   my($r_fld) = $_[0];

   # #
   $$r_fld =~ s/\x23/ /g;
   # : TO ASCII 179
   #$$r_fld =~ s/\x3A/\xB3/g;
   # ^
   #$$r_fld =~ s/\x5E/ /g;
   # |
   #$$r_fld =~ s/\x7C/ /g;
   ## @ TO ~
   #$$r_fld =~ s/\x40/\x7E/g;


   # & TO ASCII 180
   #$$r_fld =~ s/\x26/\xB4/g;

   # = TO ASCII 181
   #$$r_fld =~ s/\x3D/\xB5/g;

   # + TO ASCII 182
   #$$r_fld =~ s/\x2B/\xB6/g;

   # % TO ASCII 183
   #$$r_fld =~ s/\x25/\xB7/g;

   # " TO ASCII 184
   $$r_fld =~ s/\x22/\xB8/g;

   # ' TO ASCII 185
   $$r_fld =~ s/\x27/\xB8/g;

   # \n TO %
   $$r_fld =~ s/\n/<br>/g;

   # ASCII 13
   $$r_fld =~ s/\x0D//g;
}

sub PutBackSpCh {
   my($r_fld) = $_[0];

   # % TO \n
   #$$r_fld =~ s/\x25/\n/g;
   # ~ TO @
   #$$r_fld =~ s/\x7E/\x40/g;
   # ASCII 179 to :
   #$$r_fld =~ s/\xB3/\x3A/g;
   # ASCII 180 to &
   #$$r_fld =~ s/\xB4/\x26/g;
   # ASCII 181 to =
   #$$r_fld =~ s/\xB5/\x3D/g;
   # ASCII 182 to +
   #$$r_fld =~ s/\xB6/\x2B/g;
   # ASCII 183 to %
   #$$r_fld =~ s/\xB7/\x25/g;
   # ASCII 184 to '
   $$r_fld =~ s/\xB8/\x27/g;
   # <br>
   #$$r_fld =~ s/\xB8/\x22/g;
}

sub ChkChr {
   my($r_ck) = $_[0];

   if ($$r_ck =~ /\|/ ) {
      $$r_ck =~ s/\|//g;
   }

   if (!$$r_ck) {
      &WrError('ChkChr', "Invalid use of Character\n");
      $main::ovar->{'SystemError'} = "A system error has occurred. Please contact the webmaster.";
      $ovar->{'AdminEmail'} = $main::gvar->{'AdminEmail'};
      $$r_ck = "$main::config{'AdminHtml'}/$main::gvar->{'SysErrorPage'}";
   }

   return 1;
}

sub ChkFile {
   my($r_ck) = $_[0];

   if ($$r_ck =~ /\|/ ) {
      $$r_ck =~ s/\|//g;
   }

   if (!(-e $$r_ck)) {
      &SysError("ChkFile", "Cannot open $$r_ck");
      #&WrError('ChkFile', "$$r_ck not found\n");
      $main::ovar->{'SystemError'} = "A file system error has occurred. Please contact the webmaster.";
    #  $ovar->{'AdminEmail'} = $main::gvar->{'AdminEmail'};
     # $$r_ck = "$main::config{'AdminHtml'}/$main::gvar->{'SysErrorPage'}";
   }

   return 1;
}

sub check_url {

    # Localize the check_referer flag which determines if user is valid.     #
    my($check_referer) = 0;

    # If a referring URL was specified, for each valid referer, make sure    #
    # that a valid referring URL was passed to FormMail.                     #

    my($referer);
    if ($ENV{'HTTP_REFERER'}) {
        foreach $referer (@referers) {
            if ($debug) { print "referer=$referer\n"; }
            if ($debug) { print "referer2=$ENV{'HTTP_REFERER'}\n"; }
            if ($ENV{'HTTP_REFERER'} =~ m|https?://([^/]*)$referer|i) {
                $check_referer = 1;
                last;
            }
        }
    }
    else {
        $check_referer = 1;
    }

    # If the HTTP_REFERER was invalid, send back an error.
    if ($check_referer != 1) {
       &InitializeOut();
       print qq(<html>\n <head>\n);
       print qq(</head>\n <body>);
       print qq(<br><br><br><center><h2> Security Violation  </h2>);
       print qq(</body>\n </html>);
       exit;
    }
}

sub PutPrefix {
   my($pre) = $_[0];
   my($aft) = $_[1];

   return( join("", $pre, $aft) );
}

sub trim {
   my @out = @_;
   for (@out) {
      s/^\s+//;
      s/^\s+$//;
   }
   return wantarray ? @out : $out[0];
}

sub ChkInteger {

   my($var) = $_[0];
   my($tmpl, $r, $lDirty, $ni, @a);
   my $wd = length($var);

   $tmpl="c".$wd;

   @a=unpack($tmpl, $var);

   $ni=0;

   $lDirty=0;
   foreach $r (@a) {
      if ($r > 57 || $r < 48) {
         $lDirty=1;
         last;
      }
   }

   return $lDirty;
}

sub commify {
   my $text = reverse $_[0];
   $text =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
   return scalar reverse $text;
}

sub died {
   my($html) = $_[0];
   my($file) = $_[1];

   # Cannot open error log
   if ($html) {
       &InitializeOut();
       print qq(<html>\n <head>\n);
       print qq(</head>\n <body>);
       print qq(<br><br><br><center><h2> Cannot open $file ... </h2>);
       print qq(</body>\n </html>);
   }
   else {
      print " Cannot open error log ... \n";
   }
   exit;
}

sub SqlDtToSec {
   my($dt) = $_[0];
   if ($debug) { print "DtToSec ....\n"; }

   if ($debug) { print "dt=$dt ....\n"; }

   my ($y, $m, $d) = split(/-/, $dt);

   # Check for invalid century
   if ($y < 2001) { return 0; }


   $y = $y - 2000;
   $y = $y + 100;
   $m=$m-1;
   if ($debug) { print "y=$y ....\n"; }
   if ($debug) { print "m=$m ....\n"; }
   if ($debug) { print "d=$d ....\n"; }

   # Find Epoch secs wrt local time zone
   my $rtrnvl = timelocal(0, 0, 0, $d, $m, $y);
   return($rtrnvl);
}


sub CreateStateId {
   my($stateid) = join("",$$,$ENV{'REMOTE_ADDR'},time);
   $stateid =~ s/\.//g;
   $stateid = substr($stateid, 0, $main::gvar->{'total_len_state_id'});
   return($stateid);
}


sub SetUpTemplates {
   my($Path) = $_[0];

   my($tpl)=''; my($file)='';

   if ($debug) { print "SetUpTemplate() ...\n"; }
   foreach $tpl (@main::othertemplates) {
      my($var)='';
      $file="$Path/$tpl";
      # Set Path and file
      &ChkFile(\$file);
      if ($debug) { print "file=$file ...\n"; }
      open(DXYZ, "$file");
      while (<DXYZ>) {
         $_ = &ParseForVars($_);
         $_ = &ParseForTag($_);
         $var .= $_;
      }
      close(DXYZ);

      # Put in the XML associated array - take off extention
      my($ext) = '';
      ($tpl, $ext) = split(/\./, $tpl);
      $main::ovar->{$tpl} = $var;
   }

   if ($debug) { print "ovarfooteraa=$ovar->{'footer'}\n"; }
   if ($debug) { print "ovarfooter33=$ovar->{'xsy'}\n"; }


}

sub GetCookies {
   # Retrieve the backup cookies
   my(@allcookies) = split(/;/, $ENV{'HTTP_COOKIE'});
   my($ck) = ();  my($nm) = ''; my($cookievalue)='';
   my(%aCook) = ();


   foreach $ck (@allcookies) {
      ($nm, $cookievalue) = split(/=/, $ck);
      $nm =~  s/\x20//g;
      if (!$aCook{$nm}) {
         $aCook{$nm} = $cookievalue;
      }
   }
   return %aCook;
}

sub VerifyLoginTicket {

   # Extract information from cookie
   my %aCookies = &GetCookies();

   if ($debug) { print "verify login ticket\n"; }

   if (!$aCookies{'SessionTicket'}) {
      if ($debug) { print "session error 1\n"; }
      if ($debug) { print "msg = $main::gvar->{'sessionerrormsg1'}\n"; }
      return(0, $main::gvar->{'sessionerrormsg1'});
   }
   my(@stuff) = split(/&/, $aCookies{'SessionTicket'});

   my $ni = 0; my(%Tick) = ();
   for($ni=0; $ni<=$#stuff; $ni+=2) {
      if ($debug) { print "name=$stuff[$ni]\n"; }
      if ($debug) { print "value=$stuff[$ni+1]\n"; }
      $Tick{$stuff[$ni]} = $stuff[$ni+1];
   }

   # Verify secret key
   if (!$main::gvar->{'s_key_today'}) {
      if ($debug) { print "session error 2\n"; }
      return(0, $main::gvar->{'sessionerrormsg2'});
   }


   my($key) = join('', $main::gvar->{'s_key_today'}, $Tick{'time'}, $Tick{'username'});

   my $ctx = Digest::MD5->new;
   $ctx->add($key);
   my($hash) = $ctx->hexdigest;
#- my($hash) = MD5->hexhash($key);

   $key = join('', $main::gvar->{'s_key_today'}, $hash);

   #-$hash = MD5->hexhash($key);
   $ctx->reset;
   $ctx->add($key);
   $hash = $ctx->hexdigest;

   if ($hash ne $Tick{'hash'}) {
      if ($debug) { print "session error 3\n"; }
      return(0, $main::gvar->{'sessionerrormsg3'});
   }

   if (!$Tick{'username'}) {
      if ($debug) { print "session error 4\n"; }
      return(0, $main::gvar->{'sessionerrormsg4'});
   }

   return(1, '', $Tick{'username'}, $Tick{'time'});

}

sub GetSessionTime {

   # Extract information from cookie
   my %aCookies = &GetCookies();

   if ($debug) { print "verify login ticket\n"; }

   if (!$aCookies{'SessionTicket'}) {
      return(0);
   }
   my(@stuff) = split(/&/, $aCookies{'SessionTicket'});

   my $ni = 0; my(%Tick) = ();
   for($ni=0; $ni<=$#stuff; $ni+=2) {
      if ($debug) { print "name=$stuff[$ni]\n"; }
      if ($debug) { print "value=$stuff[$ni+1]\n"; }
      $Tick{$stuff[$ni]} = $stuff[$ni+1];
   }

   # Verify secret key
   if (!$main::gvar->{'s_key_today'}) {
      return(0);
   }


   my($key) = join('', $main::gvar->{'s_key_today'}, $Tick{'time'}, $Tick{'username'});





   my $ctx = Digest::MD5->new;
   $ctx->add($key);
   my($hash) = $ctx->hexdigest;

   #-my($hash) = MD5->hexhash($key);

   $key = join('', $main::gvar->{'s_key_today'}, $hash);

   $ctx->reset;
   $ctx->add($key);
   $hash = $ctx->hexdigest;
   #-$hash = MD5->hexhash($key);

   if ($hash ne $Tick{'hash'}) {
      return(0);
   }

   if (!$Tick{'username'}) {
      return(0);
   }

   return($Tick{'time'});

}

sub VerifyLoginTicket_Type3 {

   # Extract information from cookie
   my %aCookies = &GetCookies();

   $main::gvar->{'login_errormsg1'} = 'Admin login denied';
   $main::gvar->{'login_errormsg3'} = 'Admin login denied';
   $main::gvar->{'login_errormsg4'} = 'Admin login denied';

   if ($debug) { print "verify login ticket\n"; }


   if (!$aCookies{'i8Yw3ewW2zWTet'}) {
   if ($debug) { print "aaaa\n"; }
      return(0, $main::gvar->{'login_errormsg3'});
   }

   # Verify secret key
   if (!$main::gvar->{'s_key_today'}) {
   if ($debug) { print "bbba\n"; }
      return(0, $main::gvar->{'login_errormsg4'});
   }

   my($key) = join('', $main::gvar->{'s_key_today'}, $main::config{'adminpassword'});

   my $ctx = Digest::MD5->new;
   $ctx->add($key);
   my($hash) = $ctx->hexdigest;
   #-my($hash) = MD5->hexhash($key);

   $key = join('', $main::gvar->{'s_key_today'}, $hash);

   $ctx->reset;
   $ctx->add($key);
   $hash = $ctx->hexdigest;
   #-$hash = MD5->hexhash($key);

   if ($debug) { print "hash=$hash\n"; }
   if ($debug) { print "hash=$aCookies{'i8Yw3ewW2zWTet'}\n"; }

   if ($hash ne $aCookies{'i8Yw3ewW2zWTet'}) {
      return(0, $main::gvar->{'login_errormsg1'});
   }


   return(1, '');

}

#  program: CreatePass
#   author: David Jourard
#     date: April/98
# descript: Creates passwords from numeric, lower and upper case letters
#
#           Options
#           -------
#                 $PassMix = 0  Only Numeric
#                            1  Only lower case
#                            2  Only upper case
#                            3  Mixes between the 3
#
sub CreatePass {
   my($PassMix) = $_[0];
   my($password_size) = $_[1];

   my($no_comp)=3; my($nj)=0;
   my($ch)=''; my($roll1)=''; my($char_no)='';  my($to)='';

   my @a_components = (8, 25, 25);
   my @strt_no = (49, 65, 97);

   # Set the seed for the rand function
   srand( time() ^ ($$ + ($$ << 15)) );

   for ($nj = 1; $nj <= $password_size; $nj++) {
      if ($PassMix == 3) {
         $roll1 = int(rand 3);
      }
      else {
         $roll1 = $PassMix;
      }
      $char_no = int(rand $a_components[$roll1]) + 1;
      if ($debug) { print "char_no=$char_no\n"; }

      $ch = chr($strt_no[$roll1]+$char_no);
      $to .= $ch;
   }
   return ($to);
}

sub CrtSessionLoginTicket {
   my($user) = $_[0];

   if ($debug) { print "CrtLoginTicket ....\n"; }
   my $time = time;

   my($key) = join('', $main::gvar->{'s_key_today'}, $time, $user);

   my $ctx = Digest::MD5->new;
   $ctx->add($key);
   my($hash) = $ctx->hexdigest;
   #-my($hash) = MD5->hexhash($key);

   $key = join('', $main::gvar->{'s_key_today'}, $hash);

   $ctx->reset;
   $ctx->add($key);
   $hash = $ctx->hexdigest;
   #-$hash = MD5->hexhash($key);

   #                                   'expires' => $expires }
   return CGI::Cookie->new(-name=>'SessionTicket',
                           -path=>'/',
                           -value =>{ 'time' => $time,
                                      'username' => $user,
                                      'hash' => $hash }
                           );


}


sub CheckSessionTime {
   my $username = $_[0];
   my $sessiontime = $_[1];
   my $message = '';

      if ($debug) { print "sessiontime=$sessiontime\n"; }
   if ($main::gvar->{'DatabaseSessionTime'}) {
      my(@row) = ();
      if ( !(&main::FindRecord('user',"*","username='$username'",\@row )) ) {
         $message = "$username $main::gvar->{'loginmessage1'}";
         return (0, $message);
      }
      $sessiontime = $row[$main::cvar->{'TIMEIN'}];
   }

   #if ($debug) { print "sessiontime=$sessiontime\n"; }
   if ($sessiontime > 0) {
      # checked lapsed time
      my $timenow = time;
      my $lapseseconds = $main::gvar->{'SessionTimeInMinutes'} * 60;
      #if ($debug) { print "lapsesec=$lapseseconds\n"; }
      #if ($debug) { print "timenow=$timenow\n"; }

      if ( ($timenow - $sessiontime) > $lapseseconds ) {
         # Logged in doing things
         $message = $main::gvar->{'loginmessage6'};
         return (0, $message);
      }
   }

   return 1;

}


sub UpDtSessionTime {
   my $username = $_[0];

   if ($main::gvar->{'DatabaseSessionTime'}) {
      # Update user table with log in time
      my @datauser = ();  my %FldType = (); my %VALUES = ();
      $datauser[0] = 'timein';
      $VALUES{'timein'} = time;
      &main::UpDtData('user', \@datauser, \%VALUES, "username='$username'");
   }
   else {
      # Update session time via the login cookie
      my $cookie = &genlib::CrtSessionLoginTicket($username);
      print "Set-Cookie: $cookie\n";
   }

   return 1;
}

sub UpdtParameterFile {
   my $outputfile = $_[0];
   my $templatefile = $_[1];
   my $rh_VALUES = $_[2];

   my $var = '';  my $varvalue = '';  my %OldVar = ();  my $key = '';

   open(FL1, "$main::config{'DataPath'}/$outputfile");
   while (<FL1>) {
      chop($_);
      ($var, $varvalue) = split(/=/, $_);
      $OldVar{$var} = $varvalue;
   }
   close(FL1);

   foreach $key (keys (%{$rh_VALUES}) ) {
      # Update variables
      $OldVar{$key} = $$rh_VALUES{$key};
   }

   open(FL, ">$main::config{'DataPath'}/$outputfile");
   &genlib::RdOutFile("$main::config{'DataPath'}/$templatefile", \*FL);
   close(FL);


   # Write out the vars
   open(FL1, ">$main::config{'DataPath'}/$outputfile");
   foreach $key (keys (%OldVar) ) {
      print FL1 "$key=$OldVar{$key}\n";
   }
   close(FL1);
}


1; #return true
